/*

        CLOCK 8MHz

 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "lcdInterface.h"

volatile char timer10msFlag;
volatile int timerCounter;

#define ADC_GET_VOLT        (_BV(REFS0) | 4)
#define ADC_GET_TEMP        (_BV(REFS0) | 3)
volatile char highVoltEnableFlag;
volatile char adcStatus; 
volatile int adcCount;
volatile int adcData;
volatile long adcVolt;
volatile long adcTemp;
//#define ADC_VAL(mv)         ((((long)mv) * 1024L * 22L) / (1100L * (22L + 56L)))
#define ADC_VAL(mv)         ((((long)mv) * 1024L) / (5000L))
#define ADC_TEMP(adc)       ((adc * 5000L / 1024L) - 600L)

#define PWM_MAX             19
#define ERR_MAX             16
#define VCONT_MAX           127

volatile int vContP;
volatile int vContC;
volatile int vErrP;
volatile int vErrC;

volatile char lcdPulsWidthType = 6;
char pulsWidth[] = {
    3, 10,      // 0
    3, 6,       // 5
    3, 4,       // 10
    2, 37,      // 15
    3, 2,       // 20
    2, 23,      // 25
    3, 1,       // 30
    2, 15,      // 35
    2, 13,      // 40
    2, 11,      // 45
    3, 0        // 50
};

int getPulsWidth(char type)
{
    int result = (pulsWidth[type * 2] << 8) | pulsWidth[type * 2 + 1];
    return result;
}

void tempControl(int val)
{
    char newType;
    int temp = ADC_TEMP(val);
    if(temp < 500) {
        newType = (temp + 25) / 50;
    } else {
        newType = 10;
    }
    if(lcdPulsWidthType != newType) {
        initPulsWidth(getPulsWidth(newType));
        lcdPulsWidthType = newType;
    }
}

void highVoltEnable(char flag)
{
    highVoltEnableFlag = flag;
    if(flag) {
        PORTD &= ~_BV(PD6);
        vContP = 0;
        vErrP = 0;
        OCR0A = PWM_MAX;
        OCR0B = 0;
        TCCR0A = _BV(COM0B1) | _BV(WGM01) | _BV(WGM00);
        TCCR0B = _BV(WGM02) | _BV(CS00);
    } else {
        TCCR0A = 0;
        TCCR0B = 0;
        PORTD |= _BV(PD6);
    }
}

void voltControl(int val)
{
    int newVal;
    if(highVoltEnableFlag) {
        vErrC = ADC_VAL(3500) - val;
        if(vErrC > ERR_MAX) vErrC = ERR_MAX;
        if(vErrC < -ERR_MAX) vErrC = -ERR_MAX;
        vContC = (vErrC - (vErrP >> 1)) + vContP;
        if(vContC < 0) vContC = 0;
        if(vContC > VCONT_MAX) vContC = VCONT_MAX;
        newVal = vContC >> 1;
        if(newVal > (PWM_MAX * 3 / 8)) newVal = (PWM_MAX * 3 / 8) - 1;
        OCR0B = newVal;
        vContP = vContC;
        vErrP = vErrC;
    } else {
        vContP = 0;
        vErrP = 0;
    }
}

ISR(TIMER1_COMPA_vect)
{
    timer10msFlag = 1;
}

ISR(ADC_vect)
{
    char voltFlag;
    adcData = ADC;

    if(adcStatus == ADC_GET_VOLT) {
        adcStatus = ADC_GET_TEMP;
        voltFlag = 1;
    } else {
        adcStatus = ADC_GET_VOLT;
        voltFlag = 0;
    }
    ADMUX = adcStatus;
    ADCSRA |= _BV(ADSC);
    if(voltFlag) {
        adcVolt += adcData;
    } else {
        adcTemp += adcData;
    }
    adcCount++;
    if(adcCount >= 8) {
        adcCount = 0;
        voltControl(adcVolt/4);
        tempControl(adcTemp/4);
        adcVolt = 0;
        adcTemp = 0;
    }
}

void init()
{
// Port Setting
    PORTD = 0x40;
    DDRD = 0x68;
// High Voltage Regulator
    highVoltEnable(0);
// Timer
    OCR1A = 10000 - 1;  // 10mS
    TCCR1A = 0;
    TCCR1B = _BV(WGM12) | _BV(CS11);
    TIMSK1 |= _BV(OCIE1A);
// Clock Genarator  (2MHz)
    OCR2A = 3;
    OCR2B = 1;
    TCCR2A = _BV(COM2B1) | _BV(WGM21) | _BV(WGM20);
    TCCR2B = _BV(WGM22) | _BV(CS20);
// ADC for Regulator Voltage and Temp
    adcStatus = ADC_GET_VOLT;
    ADMUX = adcStatus;
    ADCSRA = _BV(ADEN) | _BV(ADSC) | _BV(ADIE) | _BV(ADPS2) | _BV(ADPS0);
}

void testPattarn()
{
    static char pat = 0;
    writeCmnd(0b10100000);          // 
    writeCmnd(0b00000000);          // @JnC  0C
    writeCmnd(0b11000001);          // @IC 65C
    writeCmnd(0b01000000 | 8);      // y[WAhXZbg
    writeCmnd(0b00010000);          // JAhXZbg (H)
    writeCmnd(0b00000000);          // JAhXZbgiL)
    for(int inx = 0; inx < 50; inx++) {    // 64C65CNA
        writeData(1);
    }
    for(int inx = 0; inx < 50; inx++) {    // 64C65CNA
        writeData(3);
    }
    for(int inx = 0; inx < 30; inx++) {    // 64C65CNA
        writeData(2);
    }



    for(char iny = 0; iny < 8; iny++) {
        writeCmnd(0b01000000 | iny);    // y[Wݒ
        writeCmnd(0b00010000);          // Jʒuݒ(H)
        writeCmnd(0b00000000);          // @@@@@@@@@(L)
        writeData(0x0F);                // Fill First Column
        for(char inx = 0; inx < 128;) {
            if(pat == 0) {
                writeData(0b01010101);
                writeData(0b10101010);
                inx += 2;
            } else if(pat == 1) {
                writeData(0b11001100);
                writeData(0b11001100);
                writeData(0b00110011);
                writeData(0b00110011);
                inx += 4;
            } else if(pat == 2) {
                if(((iny == 1) || (iny == 2)) && ((inx > 16) && (inx < 64))) {
                    if((inx > 32) && (inx < 48)) {
                        for(char a = 0; a < 8; a++) {
                            writeData(0);
                        }
                    } else {
                        for(char a = 0; a < 4; a++) {
                            writeData(0b10101010);
                            writeData(0b01010101);
                        }
                    }
                } else if(((iny > 3) && (iny < 7)) && ((inx > 72) && (inx < 104))) {
                    for(char a = 0; a < 2; a++) {
                        writeData(0b11001100);
                        writeData(0b11001100);
                        writeData(0b00110011);
                        writeData(0b00110011);
                    }
                } else {
                    for(char a = 0; a < 4; a++) {
                        writeData(0b11110000);
                    }
                    for(char a = 0; a < 4; a++) {
                        writeData(0b00001111);
                    }
                }
                inx += 8;
            } else if(pat == 3) {
                if(((iny == 2) || (iny == 3)) && ((inx > 32) && (inx < 64))) {
                    for(char a = 0; a < 8; a++) {
                        writeData(0b10101010);
                        writeData(0b01010101);
                    }
                } else {
                    if(iny & 0x01) {
                        for(char a = 0; a < 8; a++) {
                            writeData(0b11111111);
                        }
                    }
                    for(char a = 0; a < 8; a++) {
                        writeData(0b00000000);
                    }
                    if(!(iny & 0x01)) {
                        for(char a = 0; a < 8; a++) {
                            writeData(0b11111111);
                        }
                    }
                }
                inx += 16;
           } else if(pat == 4) {        // frame
                if(iny == 0) {
                    writeData(0xFF);
                    for(char a = 1; a < 127; a++) {
                        writeData(0x01);
                    }
                    writeData(0xFF);
                } else
                if(iny == 7) {
                    writeData(0xFF);
                    for(char a = 1; a < 127; a++) {
                        writeData(0x80);
                    }
                    writeData(0xFF);
                } else {
                    writeData(0xFF);
                    for(char a = 1; a < 127; a++) {
                        writeData(0);
                    }
                    writeData(0xFF);
                }
                inx += 130;
           } else {
                for(char a = 1; a < 129; a++) {
                    writeData(0xFF);
                }
                inx += 130;
           }
        }
        writeData(0x55);                // Fill Last Column
    }
    highVoltEnable(1);
    _delay_ms(100);
    writeCmnd(0b10010001);
    _delay_ms(100);
    while(checkBusy());
    highVoltEnable(0);
    pat++;
    if(pat > 4) pat = 0;
}

int main()
{
    CLKPR = _BV(CLKPCE);
    CLKPR = 0;
    init();
    initInterface();
    initLcd(getPulsWidth(lcdPulsWidthType));

    sei();

    while(1)
    {
        if(timer10msFlag) {
            timerCounter++;
            if(timerCounter >= 100) {  // 1 sec
                timerCounter = 0;
                testPattarn();
            }
            timer10msFlag = 0;
        }
    }
}
